"""
Excel reader and writer for AMS.

This module leverages the existing parser and writer in andes.io.xlsx.
"""
import logging

from andes.io.xlsx import (read, testlines, confirm_overwrite, _add_book)  # NOQA

from ams.shared import pd
from ams.shared import summary_row, summary_name, model2df


logger = logging.getLogger(__name__)


def write(system, outfile,
          skip_empty=True, overwrite=None, add_book=None,
          to_andes=False,
          ):
    """
    Write loaded AMS system data into an xlsx file

    Revised from `andes.io.xlsx.write`, where non-ANDES models
    are skipped if `to_andes` is True.

    Parameters
    ----------
    system : System
        A loaded system with parameters
    outfile : str
        Path to the output file
    skip_empty : bool
        Skip output of empty models (n = 0)
    overwrite : bool, optional
        None to prompt for overwrite selection; True to overwrite; False to not overwrite
    add_book : str, optional
        An optional model to be added to the output spreadsheet
    to_andes : bool, optional
        Write to an ANDES system, where non-ANDES models are skipped

    Returns
    -------
    bool
        True if file written; False otherwise
    """
    if not confirm_overwrite(outfile, overwrite=overwrite):
        return False

    writer = pd.ExcelWriter(outfile, engine='xlsxwriter')
    writer = _write_system(system, writer, skip_empty, to_andes=to_andes)
    writer = _add_book(system, writer, add_book)

    writer.close()

    logger.info('xlsx file written to "%s"', outfile)
    return True


def _write_system(system, writer, skip_empty, to_andes=False):
    """
    Write the system to pandas ExcelWriter

    Revised from `andes.io.xlsx._write_system`, where non-ANDES models
    are skipped if `to_andes` is True.
    """
    summary_found = False
    for name, instance in system.models.items():
        df = model2df(instance, skip_empty, to_andes=to_andes)
        if df is None:
            continue

        if name == summary_name:
            df = pd.concat([pd.DataFrame([summary_row]), df], ignore_index=True)
            df.index.name = "uid"
            summary_found = True

        df.to_excel(writer, sheet_name=name, freeze_panes=(1, 0))

    if not summary_found:
        df = pd.DataFrame([summary_row])
        df.index.name = "uid"
        df.to_excel(writer, sheet_name=summary_name, freeze_panes=(1, 0))

    return writer



"""
RenGen scheduling model.
"""

from andes.core.param import NumParam, IdxParam
from andes.core.model import ModelData
from ams.core.model import Model


class REGCData(ModelData):
    """
    Data container for RenGen scheduling model.
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="interface bus idx",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Sn = NumParam(default=100.0, tex_name='S_n',
                           info='device MVA rating',
                           unit='MVA',
                           )
        self.gammap = NumParam(default=1.0,
                               info="P ratio of linked static gen",
                               tex_name=r'\gamma_P'
                               )
        self.gammaq = NumParam(default=1.0,
                               info="Q ratio of linked static gen",
                               tex_name=r'\gamma_Q'
                               )


class REGCA1(REGCData, Model):
    """
    Renewable generator scheduling model.

    References
    ----------
    1. ANDES Documentation, REGCA1,
       https://andes.readthedocs.io/en/stable/groupdoc/RenGen.html#regca1
    """

    def __init__(self, system=None, config=None) -> None:
        REGCData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'RenGen'


class REGCV1(REGCData, Model):
    """
    Voltage-controlled converter model (virtual synchronous generator) with
    inertia emulation, for scheduling purposes.

    Here Mmax and Dmax are assumed to be constant, but they might subject to
    the operating condition of the converter.

    Notes
    -----
    - The generation is defined by group :ref:`StaticGen`
    - Generation cost is defined by model :ref:`GCost`
    - Inertia emulation cost is defined by model :ref:`VSGCost`

    References
    ----------
    1. ANDES Documentation, REGCV1,
       https://andes.readthedocs.io/en/stable/groupdoc/RenGen.html#regcv1
    """

    def __init__(self, system=None, config=None) -> None:
        REGCData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'VSG'
        self.M = NumParam(default=10, tex_name='M',
                          info='Inertia emulation',
                          unit='s',
                          power=True,)
        self.D = NumParam(default=0, tex_name='D',
                          info='Damping emulation',
                          unit='p.u.',
                          power=True,)
        self.Mmax = NumParam(default=99, tex_name='M_{max}',
                             info='Maximum inertia emulation',
                             unit='s',
                             power=True,)
        self.Dmax = NumParam(default=99, tex_name='D_{max}',
                             info='Maximum damping emulation',
                             unit='p.u.',
                             power=True,)


class REGCV2(REGCV1):
    """
    Alias for REGCV1, for scheduling purposes.

    References
    ----------
    1. ANDES Documentation, REGCV2,
       https://andes.readthedocs.io/en/stable/groupdoc/RenGen.html#regcv2
    """

    def __init__(self, system=None, config=None) -> None:
        REGCV1.__init__(self, system, config)



import unittest

import os

import ams
from ams.utils.paths import list_cases, get_export_path


class TestPaths(unittest.TestCase):
    def setUp(self) -> None:
        self.npcc = 'npcc/'
        self.matpower = 'matpower/'
        self.ieee14 = ams.get_case("ieee14/ieee14.raw")

    def test_tree(self):
        list_cases(self.npcc, no_print=True)
        list_cases(self.matpower, no_print=True)

    def test_relative_path(self):
        ss = ams.run('ieee14.raw',
                     input_path=ams.get_case('ieee14/', check=False),
                     no_output=True, default_config=True,
                     )
        self.assertNotEqual(ss, None)


class TestExportPath(unittest.TestCase):

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case('5bus/pjm5bus_demo.json'),
                           setup=True, no_output=True)

    def test_no_fullname(self):
        """
        Test export path with no full name specified.
        """
        self.ss.files.full_name = None

        path, file_name = get_export_path(self.ss,
                                          'DCOPF',
                                          path=None,
                                          fmt='csv')

        dir_path, only_file_name = os.path.split(path)
        self.assertTrue(os.path.exists(dir_path))
        self.assertIsNotNone(only_file_name)
        self.assertEqual(only_file_name, file_name)

    def test_no_path(self):
        """
        Test export path with no path specified.
        """
        path, file_name = get_export_path(self.ss,
                                          'DCOPF',
                                          path=None,
                                          fmt='csv')

        dir_path, only_file_name = os.path.split(path)
        self.assertTrue(os.path.exists(dir_path))
        self.assertIsNotNone(only_file_name)
        self.assertEqual(only_file_name, file_name)

    def test_current_path(self):
        """
        Test export path with current path specified.
        """
        path, file_name = get_export_path(self.ss,
                                          'DCOPF',
                                          path='.',
                                          fmt='csv')

        dir_path, only_file_name = os.path.split(path)
        self.assertTrue(os.path.exists(dir_path))
        self.assertIsNotNone(only_file_name)
        self.assertEqual(only_file_name, file_name)

    def test_path_with_file_name(self):
        """
        Test export path with path and file name specified.
        """
        path, file_name = get_export_path(self.ss,
                                          'DCOPF',
                                          path='./test_export.csv',
                                          fmt='csv',)

        dir_path, only_file_name = os.path.split(path)
        self.assertTrue(os.path.exists(dir_path))
        self.assertEqual(only_file_name, 'test_export.csv')
        self.assertEqual(only_file_name, file_name)
